@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import json
@@ -7,13 +7,11 @@ from contextlib import contextmanager, suppress
 from functools import partial
 from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple
 
-from .conf.utils import BadLine, load_config as _load_config, parse_config_base
+from .conf.utils import BadLine, parse_config_base
+from .conf.utils import load_config as _load_config
 from .constants import cache_dir, defconf
 from .options.types import Options, defaults, option_names
-from .options.utils import (
-    KeyDefinition, KeyMap, MouseMap, MouseMapping, SequenceMap,
-    build_action_aliases
-)
+from .options.utils import KeyboardMode, KeyboardModeMap, KeyDefinition, MouseMap, MouseMapping, build_action_aliases
 from .typing import TypedDict
 from .utils import log_error
 
@@ -49,13 +47,12 @@ def atomic_save(data: bytes, path: str) -> None:
         except FileNotFoundError:
             pass
         except Exception as err:
-            log_error('Failed to delete temp file {} for atomic save with error: {}'.format(
-                p, err))
+            log_error(f'Failed to delete temp file {p} for atomic save with error: {err}')
 
 
 @contextmanager
 def cached_values_for(name: str) -> Generator[Dict[str, Any], None, None]:
-    cached_path = os.path.join(cache_dir(), name + '.json')
+    cached_path = os.path.join(cache_dir(), f'{name}.json')
     cached_values: Dict[str, Any] = {}
     try:
         with open(cached_path, 'rb') as f:
@@ -63,8 +60,7 @@ def cached_values_for(name: str) -> Generator[Dict[str, Any], None, None]:
     except FileNotFoundError:
         pass
     except Exception as err:
-        log_error('Failed to load cached in {} values with error: {}'.format(
-            name, err))
+        log_error(f'Failed to load cached in {name} values with error: {err}')
 
     yield cached_values
 
@@ -72,8 +68,7 @@ def cached_values_for(name: str) -> Generator[Dict[str, Any], None, None]:
         data = json.dumps(cached_values).encode('utf-8')
         atomic_save(data, cached_path)
     except Exception as err:
-        log_error('Failed to save cached values with error: {}'.format(
-            err))
+        log_error(f'Failed to save cached values with error: {err}')
 
 
 def commented_out_default_config() -> str:
@@ -105,28 +100,26 @@ def finalize_keys(opts: Options, accumulate_bad_lines: Optional[List[BadLine]] =
                 else:
                     accumulate_bad_lines.append(BadLine(d.definition_location.number, d.definition_location.line, err, d.definition_location.file))
 
-    keymap: KeyMap = {}
-    sequence_map: SequenceMap = {}
+    modes: KeyboardModeMap = {'': KeyboardMode()}
 
     for defn in defns:
-        is_no_op = defn.is_no_op
-        if defn.is_sequence:
-            keymap.pop(defn.trigger, None)
-            s = sequence_map.setdefault(defn.trigger, {})
-            if is_no_op:
-                s.pop(defn.rest, None)
-                if not s:
-                    del sequence_map[defn.trigger]
+        if defn.options.new_mode:
+            modes[defn.options.new_mode] = nm = KeyboardMode(defn.options.new_mode)
+            nm.on_unknown = defn.options.on_unknown
+            nm.on_action = defn.options.on_action
+            defn.definition = f'push_keyboard_mode {defn.options.new_mode}'
+        try:
+            m = modes[defn.options.mode]
+        except KeyError:
+            kerr = f'The keyboard mode {defn.options.mode} is unknown, ignoring the mapping'
+            if accumulate_bad_lines is None:
+                log_error(kerr)
             else:
-                s[defn.rest] = defn.definition
-        else:
-            sequence_map.pop(defn.trigger, None)
-            if is_no_op:
-                keymap.pop(defn.trigger, None)
-            else:
-                keymap[defn.trigger] = defn.definition
-    opts.keymap = keymap
-    opts.sequence_map = sequence_map
+                dl = defn.definition_location
+                accumulate_bad_lines.append(BadLine(dl.number, dl.line, KeyError(kerr), dl.file))
+            continue
+        m.keymap[defn.trigger].append(defn)
+    opts.keyboard_modes = modes
 
 
 def finalize_mouse_mappings(opts: Options, accumulate_bad_lines: Optional[List[BadLine]] = None) -> None:
@@ -145,11 +138,10 @@ def finalize_mouse_mappings(opts: Options, accumulate_bad_lines: Optional[List[B
     mousemap: MouseMap = {}
 
     for defn in defns:
-        is_no_op = defn.is_no_op
-        if is_no_op:
-            mousemap.pop(defn.trigger, None)
-        else:
+        if defn.definition:
             mousemap[defn.trigger] = defn.definition
+        else:
+            mousemap.pop(defn.trigger, None)
     opts.mousemap = mousemap
 
 
@@ -169,7 +161,8 @@ def load_config(*paths: str, overrides: Optional[Iterable[str]] = None, accumula
     from .options.parse import merge_result_dicts
 
     overrides = tuple(overrides) if overrides is not None else ()
-    opts_dict, paths = _load_config(defaults, partial(parse_config, accumulate_bad_lines=accumulate_bad_lines), merge_result_dicts, *paths, overrides=overrides)
+    opts_dict, found_paths = _load_config(
+        defaults, partial(parse_config, accumulate_bad_lines=accumulate_bad_lines), merge_result_dicts, *paths, overrides=overrides)
     opts = Options(opts_dict)
 
     opts.alias_map = build_action_aliases(opts.kitten_alias, 'kitten')
@@ -184,7 +177,8 @@ def load_config(*paths: str, overrides: Optional[Iterable[str]] = None, accumula
     if opts.background_opacity < 1.0 and opts.macos_titlebar_color > 0:
         log_error('Cannot use both macos_titlebar_color and background_opacity')
         opts.macos_titlebar_color = 0
-    opts.config_paths = paths
+    opts.config_paths = found_paths
+    opts.all_config_paths = paths
     opts.config_overrides = overrides
     return opts
 
