@@ -6,7 +6,6 @@
 
 #include "state.h"
 #include "cleanup.h"
-#include "fonts.h"
 #include "monotonic.h"
 #include "charsets.h"
 #include <structmember.h>
@@ -33,7 +32,12 @@ extern size_t cocoa_get_workspace_ids(void *w, size_t *workspace_ids, size_t arr
 extern monotonic_t cocoa_cursor_blink_interval(void);
 
 
-static GLFWcursor *standard_cursor = NULL, *click_cursor = NULL, *arrow_cursor = NULL;
+typedef struct mouse_cursor {
+    GLFWcursor *glfw;
+    bool initialized, is_custom;
+} mouse_cursor;
+
+static mouse_cursor cursors[GLFW_INVALID_CURSOR+1] = {0};
 
 static void set_os_window_dpi(OSWindow *w);
 
@@ -112,7 +116,7 @@ min_size_for_os_window(OSWindow *window, int *min_width, int *min_height) {
 }
 
 
-static void
+void
 adjust_window_size_for_csd(OSWindow *w, int width, int height, int *adjusted_width, int *adjusted_height) {
     *adjusted_width = width; *adjusted_height = height;
     if (global_state.is_wayland) {
@@ -318,9 +322,13 @@ change_live_resize_state(OSWindow *w, bool in_progress) {
     if (in_progress != w->live_resize.in_progress) {
         w->live_resize.in_progress = in_progress;
         w->live_resize.num_of_resize_events = 0;
-        apply_swap_interval(in_progress ? 0 : -1);
 #ifdef __APPLE__
         cocoa_out_of_sequence_render(w);
+#else
+        GLFWwindow *orig_ctx = make_os_window_context_current(w);
+        apply_swap_interval(in_progress ? 0 : -1);
+        if (orig_ctx) glfwMakeContextCurrent(orig_ctx);
+
 #endif
     }
 }
@@ -382,25 +390,31 @@ refresh_callback(GLFWwindow *w) {
 
 static int mods_at_last_key_or_button_event = 0;
 
+#ifndef __APPLE__
+typedef struct modifier_key_state {
+    bool left, right;
+} modifier_key_state;
+
 static int
-key_to_modifier(uint32_t key) {
+key_to_modifier(uint32_t key, bool *is_left) {
+    *is_left = false;
     switch(key) {
-        case GLFW_FKEY_LEFT_SHIFT:
+        case GLFW_FKEY_LEFT_SHIFT: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_SHIFT:
             return GLFW_MOD_SHIFT;
-        case GLFW_FKEY_LEFT_CONTROL:
+        case GLFW_FKEY_LEFT_CONTROL: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_CONTROL:
             return GLFW_MOD_CONTROL;
-        case GLFW_FKEY_LEFT_ALT:
+        case GLFW_FKEY_LEFT_ALT: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_ALT:
             return GLFW_MOD_ALT;
-        case GLFW_FKEY_LEFT_SUPER:
+        case GLFW_FKEY_LEFT_SUPER: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_SUPER:
             return GLFW_MOD_SUPER;
-        case GLFW_FKEY_LEFT_HYPER:
+        case GLFW_FKEY_LEFT_HYPER: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_HYPER:
             return GLFW_MOD_HYPER;
-        case GLFW_FKEY_LEFT_META:
+        case GLFW_FKEY_LEFT_META: *is_left = true; /* fallthrough */
         case GLFW_FKEY_RIGHT_META:
             return GLFW_MOD_META;
         default:
@@ -408,18 +422,40 @@ key_to_modifier(uint32_t key) {
     }
 }
 
+
+static void
+update_modifier_state_on_modifier_key_event(GLFWkeyevent *ev, int key_modifier, bool is_left) {
+    // Update mods state to be what the kitty keyboard protocol requires, as on Linux modifier key events do not update modifier bits
+    static modifier_key_state all_states[8] = {0};
+    modifier_key_state *state = all_states + MIN((unsigned)__builtin_ctz(key_modifier), sizeof(all_states)-1);
+    const int modifier_was_set_before_event = ev->mods & key_modifier;
+    const bool is_release = ev->action == GLFW_RELEASE;
+    if (modifier_was_set_before_event) {
+        // a press with modifier already set means other modifier key is pressed
+        if (!is_release) { if (is_left) state->right = true; else state->left = true;  }
+    } else {
+        // if modifier is not set before event, means both keys are released
+        state->left = false; state->right = false;
+    }
+    if (is_release) {
+        if (is_left) state->left = false; else state->right = false;
+        if (modifier_was_set_before_event && !state->left && !state->right) ev->mods &= ~key_modifier;
+    } else {
+        if (is_left) state->left = true; else state->right = true;
+        ev->mods |= key_modifier;
+    }
+}
+#endif
+
 static void
 key_callback(GLFWwindow *w, GLFWkeyevent *ev) {
     if (!set_callback_window(w)) return;
+#ifndef __APPLE__
+    bool is_left;
+    int key_modifier = key_to_modifier(ev->key, &is_left);
+    if (key_modifier != -1) update_modifier_state_on_modifier_key_event(ev, key_modifier, is_left);
+#endif
     mods_at_last_key_or_button_event = ev->mods;
-    int key_modifier = key_to_modifier(ev->key);
-    if (key_modifier != -1) {
-        if (ev->action == GLFW_RELEASE) {
-            mods_at_last_key_or_button_event &= ~key_modifier;
-        } else {
-            mods_at_last_key_or_button_event |= key_modifier;
-        }
-    }
     global_state.callback_os_window->cursor_blink_zero_time = monotonic();
     if (is_window_ready_for_callbacks()) on_key_input(ev);
     global_state.callback_os_window = NULL;
@@ -675,20 +711,53 @@ draw_single_ascii_char(const char ch, size_t *result_width, size_t *result_heigh
 #endif
 // }}}
 
+static void
+set_glfw_mouse_cursor(GLFWwindow *w, GLFWCursorShape shape) {
+    if (!cursors[shape].initialized) {
+        cursors[shape].initialized = true;
+        cursors[shape].glfw = glfwCreateStandardCursor(shape);
+    }
+    if (cursors[shape].glfw) glfwSetCursor(w, cursors[shape].glfw);
+}
+
 void
 set_mouse_cursor(MouseShape type) {
     if (global_state.callback_os_window) {
         GLFWwindow *w = (GLFWwindow*)global_state.callback_os_window->handle;
         switch(type) {
-            case HAND:
-                glfwSetCursor(w, click_cursor);
-                break;
-            case ARROW:
-                glfwSetCursor(w, arrow_cursor);
-                break;
-            default:
-                glfwSetCursor(w, standard_cursor);
-                break;
+            case INVALID_POINTER: break;
+            /* start enum to glfw (auto generated by gen-key-constants.py do not edit) */
+        case DEFAULT_POINTER: set_glfw_mouse_cursor(w, GLFW_DEFAULT_CURSOR); break;
+        case TEXT_POINTER: set_glfw_mouse_cursor(w, GLFW_TEXT_CURSOR); break;
+        case POINTER_POINTER: set_glfw_mouse_cursor(w, GLFW_POINTER_CURSOR); break;
+        case HELP_POINTER: set_glfw_mouse_cursor(w, GLFW_HELP_CURSOR); break;
+        case WAIT_POINTER: set_glfw_mouse_cursor(w, GLFW_WAIT_CURSOR); break;
+        case PROGRESS_POINTER: set_glfw_mouse_cursor(w, GLFW_PROGRESS_CURSOR); break;
+        case CROSSHAIR_POINTER: set_glfw_mouse_cursor(w, GLFW_CROSSHAIR_CURSOR); break;
+        case CELL_POINTER: set_glfw_mouse_cursor(w, GLFW_CELL_CURSOR); break;
+        case VERTICAL_TEXT_POINTER: set_glfw_mouse_cursor(w, GLFW_VERTICAL_TEXT_CURSOR); break;
+        case MOVE_POINTER: set_glfw_mouse_cursor(w, GLFW_MOVE_CURSOR); break;
+        case E_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_E_RESIZE_CURSOR); break;
+        case NE_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_NE_RESIZE_CURSOR); break;
+        case NW_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_NW_RESIZE_CURSOR); break;
+        case N_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_N_RESIZE_CURSOR); break;
+        case SE_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_SE_RESIZE_CURSOR); break;
+        case SW_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_SW_RESIZE_CURSOR); break;
+        case S_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_S_RESIZE_CURSOR); break;
+        case W_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_W_RESIZE_CURSOR); break;
+        case EW_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_EW_RESIZE_CURSOR); break;
+        case NS_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_NS_RESIZE_CURSOR); break;
+        case NESW_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_NESW_RESIZE_CURSOR); break;
+        case NWSE_RESIZE_POINTER: set_glfw_mouse_cursor(w, GLFW_NWSE_RESIZE_CURSOR); break;
+        case ZOOM_IN_POINTER: set_glfw_mouse_cursor(w, GLFW_ZOOM_IN_CURSOR); break;
+        case ZOOM_OUT_POINTER: set_glfw_mouse_cursor(w, GLFW_ZOOM_OUT_CURSOR); break;
+        case ALIAS_POINTER: set_glfw_mouse_cursor(w, GLFW_ALIAS_CURSOR); break;
+        case COPY_POINTER: set_glfw_mouse_cursor(w, GLFW_COPY_CURSOR); break;
+        case NOT_ALLOWED_POINTER: set_glfw_mouse_cursor(w, GLFW_NOT_ALLOWED_CURSOR); break;
+        case NO_DROP_POINTER: set_glfw_mouse_cursor(w, GLFW_NO_DROP_CURSOR); break;
+        case GRAB_POINTER: set_glfw_mouse_cursor(w, GLFW_GRAB_CURSOR); break;
+        case GRABBING_POINTER: set_glfw_mouse_cursor(w, GLFW_GRABBING_CURSOR); break;
+/* end enum to glfw */
         }
     }
 }
@@ -715,13 +784,14 @@ set_default_window_icon(PyObject UNUSED *self, PyObject *args) {
 }
 
 
-void
+void*
 make_os_window_context_current(OSWindow *w) {
     GLFWwindow *current_context = glfwGetCurrentContext();
     if (w->handle != current_context) {
         glfwMakeContextCurrent(w->handle);
-        global_state.current_opengl_context_id = w->id;
+        return current_context;
     }
+    return NULL;
 }
 
 void
@@ -735,6 +805,16 @@ set_os_window_size(OSWindow *os_window, int x, int y) {
     glfwSetWindowSize(os_window->handle, x, y);
 }
 
+void
+get_os_window_pos(OSWindow *os_window, int *x, int *y) {
+    glfwGetWindowPos(os_window->handle, x, y);
+}
+
+void
+set_os_window_pos(OSWindow *os_window, int x, int y) {
+    glfwSetWindowPos(os_window->handle, x, y);
+}
+
 static void
 get_window_content_scale(GLFWwindow *w, float *xscale, float *yscale, double *xdpi, double *ydpi) {
     // if you change this function also change createSurface() in wl_window.c
@@ -970,13 +1050,15 @@ native_window_handle(GLFWwindow *w) {
 
 static PyObject*
 create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
-    int x = -1, y = -1, window_state = WINDOW_NORMAL, disallow_override_title = 0;
+    int x = INT_MIN, y = INT_MIN, window_state = WINDOW_NORMAL, disallow_override_title = 0;
     char *title, *wm_class_class, *wm_class_name;
-    PyObject *optional_window_state = NULL, *load_programs = NULL, *get_window_size, *pre_show_callback;
+    PyObject *optional_window_state = NULL, *load_programs = NULL, *get_window_size, *pre_show_callback, *optional_x = NULL, *optional_y = NULL;
     static const char* kwlist[] = {"get_window_size", "pre_show_callback", "title", "wm_class_name", "wm_class_class", "window_state", "load_programs", "x", "y", "disallow_override_title", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kw, "OOsss|OOiip", (char**)kwlist,
-        &get_window_size, &pre_show_callback, &title, &wm_class_name, &wm_class_class, &optional_window_state, &load_programs, &x, &y, &disallow_override_title)) return NULL;
-    if (optional_window_state && optional_window_state != Py_None) window_state = (int) PyLong_AsLong(optional_window_state);
+    if (!PyArg_ParseTupleAndKeywords(args, kw, "OOsss|OOOOp", (char**)kwlist,
+        &get_window_size, &pre_show_callback, &title, &wm_class_name, &wm_class_class, &optional_window_state, &load_programs, &optional_x, &optional_y, &disallow_override_title)) return NULL;
+    if (optional_window_state && optional_window_state != Py_None) { if (!PyLong_Check(optional_window_state)) { PyErr_SetString(PyExc_TypeError, "window_state must be an int"); return NULL; } window_state = (int) PyLong_AsLong(optional_window_state); }
+    if (optional_x && optional_x != Py_None) { if (!PyLong_Check(optional_x)) { PyErr_SetString(PyExc_TypeError, "x must be an int"); return NULL;} x = (int)PyLong_AsLong(optional_x); }
+    if (optional_y && optional_y != Py_None) { if (!PyLong_Check(optional_y)) { PyErr_SetString(PyExc_TypeError, "y must be an int"); return NULL;} y = (int)PyLong_AsLong(optional_y); }
     if (window_state < WINDOW_NORMAL || window_state > WINDOW_MINIMIZED) window_state = WINDOW_NORMAL;
 
     static bool is_first_window = true;
@@ -992,6 +1074,8 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
         glfwSetHasCurrentSelectionCallback(has_current_selection);
         glfwSetIMECursorPositionCallback(get_ime_cursor_position);
         glfwSetSystemColorThemeChangeCallback(on_system_color_scheme_change);
+        // Request SRGB output buffer
+        glfwWindowHint(GLFW_SRGB_CAPABLE, true);
 #ifdef __APPLE__
         cocoa_set_activation_policy(OPT(macos_hide_from_tasks));
         glfwWindowHint(GLFW_COCOA_GRAPHICS_SWITCHING, true);
@@ -1072,7 +1156,7 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
     PyObject *pret = PyObject_CallFunction(pre_show_callback, "N", native_window_handle(glfw_window));
     if (pret == NULL) return NULL;
     Py_DECREF(pret);
-    if (x != -1 && y != -1) glfwSetWindowPos(glfw_window, x, y);
+    if (x != INT_MIN && y != INT_MIN) glfwSetWindowPos(glfw_window, x, y);
 #ifndef __APPLE__
     glfwShowWindow(glfw_window);
 #endif
@@ -1090,15 +1174,12 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
         PyObject *ret = PyObject_CallFunction(load_programs, "O", is_semi_transparent ? Py_True : Py_False);
         if (ret == NULL) return NULL;
         Py_DECREF(ret);
-#define CC(dest, shape) {\
-    if (!dest##_cursor) { \
-        dest##_cursor = glfwCreateStandardCursor(GLFW_##shape##_CURSOR); \
-        if (dest##_cursor == NULL) { log_error("Failed to create the %s mouse cursor, using default cursor.", #shape); } \
-}}
-        CC(standard, IBEAM); CC(click, HAND); CC(arrow, ARROW);
-#undef CC
         get_platform_dependent_config_values(glfw_window);
+        GLint encoding;
+        glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER, GL_BACK_LEFT, GL_FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING, &encoding);
+        if (encoding != GL_SRGB) log_error("The output buffer does not support sRGB color encoding, colors will be incorrect.");
         is_first_window = false;
+
     }
     OSWindow *w = add_os_window();
     w->handle = glfw_window;
@@ -1121,7 +1202,7 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
 #endif
     send_prerendered_sprites_for_window(w);
     if (logo.pixels && logo.width && logo.height) glfwSetWindowIcon(glfw_window, 1, &logo);
-    glfwSetCursor(glfw_window, standard_cursor);
+    set_glfw_mouse_cursor(glfw_window, GLFW_TEXT_CURSOR);
     update_os_window_viewport(w, false);
     glfwSetWindowPosCallback(glfw_window, window_pos_callback);
     // missing size callback
@@ -1145,7 +1226,6 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
     w->cursor_blink_zero_time = now;
     w->last_mouse_activity_at = now;
     w->is_semi_transparent = is_semi_transparent;
-    global_state.current_opengl_context_id = w->id;
     if (want_semi_transparent && !w->is_semi_transparent) {
         static bool warned = false;
         if (!warned) {
@@ -1310,6 +1390,12 @@ glfw_init(PyObject UNUSED *self, PyObject *args) {
 
 static PyObject*
 glfw_terminate(PYNOARG) {
+    for (size_t i = 0; i < arraysz(cursors); i++) {
+        if (cursors[i].is_custom && cursors[i].glfw) {
+            glfwDestroyCursor(cursors[i].glfw);
+            cursors[i] = (mouse_cursor){0};
+        }
+    }
     glfwTerminate();
     Py_RETURN_NONE;
 }
@@ -1612,6 +1698,10 @@ request_window_attention(id_type kitty_window_id, bool audio_bell) {
 
 void
 set_os_window_title(OSWindow *w, const char *title) {
+    if (!title) {
+        if (global_state.is_wayland) glfwWaylandRedrawCSDWindowTitle(w->handle);
+        return;
+    }
     static char buf[2048];
     strip_csi_(title, buf, arraysz(buf));
     glfwSetWindowTitle(w->handle, buf);
@@ -1692,13 +1782,108 @@ cocoa_window_id(PyObject UNUSED *self, PyObject *os_wid) {
 #endif
 }
 
+static GLFWCursorShape
+pointer_name_to_glfw_name(const char *name) {
+    /* start name to glfw (auto generated by gen-key-constants.py do not edit) */
+    if (strcmp(name, "arrow") == 0) return GLFW_DEFAULT_CURSOR;
+    if (strcmp(name, "beam") == 0) return GLFW_TEXT_CURSOR;
+    if (strcmp(name, "text") == 0) return GLFW_TEXT_CURSOR;
+    if (strcmp(name, "pointer") == 0) return GLFW_POINTER_CURSOR;
+    if (strcmp(name, "hand") == 0) return GLFW_POINTER_CURSOR;
+    if (strcmp(name, "help") == 0) return GLFW_HELP_CURSOR;
+    if (strcmp(name, "wait") == 0) return GLFW_WAIT_CURSOR;
+    if (strcmp(name, "progress") == 0) return GLFW_PROGRESS_CURSOR;
+    if (strcmp(name, "crosshair") == 0) return GLFW_CROSSHAIR_CURSOR;
+    if (strcmp(name, "cell") == 0) return GLFW_CELL_CURSOR;
+    if (strcmp(name, "vertical-text") == 0) return GLFW_VERTICAL_TEXT_CURSOR;
+    if (strcmp(name, "move") == 0) return GLFW_MOVE_CURSOR;
+    if (strcmp(name, "e-resize") == 0) return GLFW_E_RESIZE_CURSOR;
+    if (strcmp(name, "ne-resize") == 0) return GLFW_NE_RESIZE_CURSOR;
+    if (strcmp(name, "nw-resize") == 0) return GLFW_NW_RESIZE_CURSOR;
+    if (strcmp(name, "n-resize") == 0) return GLFW_N_RESIZE_CURSOR;
+    if (strcmp(name, "se-resize") == 0) return GLFW_SE_RESIZE_CURSOR;
+    if (strcmp(name, "sw-resize") == 0) return GLFW_SW_RESIZE_CURSOR;
+    if (strcmp(name, "s-resize") == 0) return GLFW_S_RESIZE_CURSOR;
+    if (strcmp(name, "w-resize") == 0) return GLFW_W_RESIZE_CURSOR;
+    if (strcmp(name, "ew-resize") == 0) return GLFW_EW_RESIZE_CURSOR;
+    if (strcmp(name, "ns-resize") == 0) return GLFW_NS_RESIZE_CURSOR;
+    if (strcmp(name, "nesw-resize") == 0) return GLFW_NESW_RESIZE_CURSOR;
+    if (strcmp(name, "nwse-resize") == 0) return GLFW_NWSE_RESIZE_CURSOR;
+    if (strcmp(name, "zoom-in") == 0) return GLFW_ZOOM_IN_CURSOR;
+    if (strcmp(name, "zoom-out") == 0) return GLFW_ZOOM_OUT_CURSOR;
+    if (strcmp(name, "alias") == 0) return GLFW_ALIAS_CURSOR;
+    if (strcmp(name, "copy") == 0) return GLFW_COPY_CURSOR;
+    if (strcmp(name, "not-allowed") == 0) return GLFW_NOT_ALLOWED_CURSOR;
+    if (strcmp(name, "no-drop") == 0) return GLFW_NO_DROP_CURSOR;
+    if (strcmp(name, "grab") == 0) return GLFW_GRAB_CURSOR;
+    if (strcmp(name, "grabbing") == 0) return GLFW_GRABBING_CURSOR;
+/* end name to glfw */
+    return GLFW_INVALID_CURSOR;
+}
+
+static PyObject*
+is_css_pointer_name_valid(PyObject *self UNUSED, PyObject *name) {
+    if (!PyUnicode_Check(name)) { PyErr_SetString(PyExc_TypeError, "pointer name must be a string"); return NULL; }
+    const char *q = PyUnicode_AsUTF8(name);
+    if (strcmp(q, "default") == 0) { Py_RETURN_TRUE; }
+    if (pointer_name_to_glfw_name(q) == GLFW_INVALID_CURSOR) { Py_RETURN_FALSE; }
+    Py_RETURN_TRUE;
+}
+
+static const char*
+glfw_name_to_css_pointer_name(GLFWCursorShape q) {
+    switch(q) {
+        case GLFW_INVALID_CURSOR: return "";
+        /* start glfw to css (auto generated by gen-key-constants.py do not edit) */
+        case GLFW_DEFAULT_CURSOR: return "default";
+        case GLFW_TEXT_CURSOR: return "text";
+        case GLFW_POINTER_CURSOR: return "pointer";
+        case GLFW_HELP_CURSOR: return "help";
+        case GLFW_WAIT_CURSOR: return "wait";
+        case GLFW_PROGRESS_CURSOR: return "progress";
+        case GLFW_CROSSHAIR_CURSOR: return "crosshair";
+        case GLFW_CELL_CURSOR: return "cell";
+        case GLFW_VERTICAL_TEXT_CURSOR: return "vertical-text";
+        case GLFW_MOVE_CURSOR: return "move";
+        case GLFW_E_RESIZE_CURSOR: return "e-resize";
+        case GLFW_NE_RESIZE_CURSOR: return "ne-resize";
+        case GLFW_NW_RESIZE_CURSOR: return "nw-resize";
+        case GLFW_N_RESIZE_CURSOR: return "n-resize";
+        case GLFW_SE_RESIZE_CURSOR: return "se-resize";
+        case GLFW_SW_RESIZE_CURSOR: return "sw-resize";
+        case GLFW_S_RESIZE_CURSOR: return "s-resize";
+        case GLFW_W_RESIZE_CURSOR: return "w-resize";
+        case GLFW_EW_RESIZE_CURSOR: return "ew-resize";
+        case GLFW_NS_RESIZE_CURSOR: return "ns-resize";
+        case GLFW_NESW_RESIZE_CURSOR: return "nesw-resize";
+        case GLFW_NWSE_RESIZE_CURSOR: return "nwse-resize";
+        case GLFW_ZOOM_IN_CURSOR: return "zoom-in";
+        case GLFW_ZOOM_OUT_CURSOR: return "zoom-out";
+        case GLFW_ALIAS_CURSOR: return "alias";
+        case GLFW_COPY_CURSOR: return "copy";
+        case GLFW_NOT_ALLOWED_CURSOR: return "not-allowed";
+        case GLFW_NO_DROP_CURSOR: return "no-drop";
+        case GLFW_GRAB_CURSOR: return "grab";
+        case GLFW_GRABBING_CURSOR: return "grabbing";
+/* end glfw to css */
+    }
+    return "";
+}
+
+static PyObject*
+pointer_name_to_css_name(PyObject *self UNUSED, PyObject *name) {
+    if (!PyUnicode_Check(name)) { PyErr_SetString(PyExc_TypeError, "pointer name must be a string"); return NULL; }
+    GLFWCursorShape s = pointer_name_to_glfw_name(PyUnicode_AsUTF8(name));
+    return PyUnicode_FromString(glfw_name_to_css_pointer_name(s));
+}
+
 static PyObject*
 set_custom_cursor(PyObject *self UNUSED, PyObject *args) {
-    int shape;
     int x=0, y=0;
     Py_ssize_t sz;
     PyObject *images;
-    if (!PyArg_ParseTuple(args, "iO!|ii", &shape, &PyTuple_Type, &images, &x, &y)) return NULL;
+    const char *shape;
+    if (!PyArg_ParseTuple(args, "sO!|ii", &shape, &PyTuple_Type, &images, &x, &y)) return NULL;
     static GLFWimage gimages[16] = {{0}};
     size_t count = MIN((size_t)PyTuple_GET_SIZE(images), arraysz(gimages));
     for (size_t i = 0; i < count; i++) {
@@ -1708,20 +1893,14 @@ set_custom_cursor(PyObject *self UNUSED, PyObject *args) {
             return NULL;
         }
     }
-#define CASE(which, dest) {\
-    case which: \
-        dest = glfwCreateCursor(gimages, x, y, count); \
-        if (dest == NULL) { PyErr_SetString(PyExc_ValueError, "Failed to create custom cursor"); return NULL; } \
-        break; \
-}
-    switch(shape) {
-        CASE(GLFW_IBEAM_CURSOR, standard_cursor);
-        CASE(GLFW_HAND_CURSOR, click_cursor);
-        CASE(GLFW_ARROW_CURSOR, arrow_cursor);
-        default:
-            PyErr_SetString(PyExc_ValueError, "Unknown cursor shape");
-            return NULL;
+    GLFWCursorShape gshape = pointer_name_to_glfw_name(shape);
+    if (gshape == GLFW_INVALID_CURSOR) { PyErr_Format(PyExc_KeyError, "Unknown pointer shape: %s", shape); return NULL; }
+    GLFWcursor *c = glfwCreateCursor(gimages, x, y, count);
+    if (c == NULL) { PyErr_SetString(PyExc_ValueError, "Failed to create custom cursor from specified images"); return NULL; }
+    if (cursors[gshape].initialized && cursors[gshape].is_custom && cursors[gshape].glfw) {
+        glfwDestroyCursor(cursors[gshape].glfw);
     }
+    cursors[gshape].initialized = true; cursors[gshape].is_custom = true; cursors[gshape].glfw = c;
     Py_RETURN_NONE;
 }
 
@@ -1948,6 +2127,8 @@ make_x11_window_a_dock_window(PyObject *self UNUSED, PyObject *args UNUSED) {
 
 static PyMethodDef module_methods[] = {
     METHODB(set_custom_cursor, METH_VARARGS),
+    METHODB(is_css_pointer_name_valid, METH_O),
+    METHODB(pointer_name_to_css_name, METH_O),
     {"create_os_window", (PyCFunction)(void (*) (void))(create_os_window), METH_VARARGS | METH_KEYWORDS, NULL},
     METHODB(set_default_window_icon, METH_VARARGS),
     METHODB(set_clipboard_data_types, METH_VARARGS),
@@ -2002,7 +2183,6 @@ init_glfw(PyObject *m) {
     ADDC(GLFW_PRESS);
     ADDC(GLFW_REPEAT);
     ADDC(true); ADDC(false);
-    ADDC(GLFW_IBEAM_CURSOR); ADDC(GLFW_HAND_CURSOR); ADDC(GLFW_ARROW_CURSOR);
     ADDC(GLFW_PRIMARY_SELECTION); ADDC(GLFW_CLIPBOARD);
 
     /* start glfw functional keys (auto generated by gen-key-constants.py do not edit) */
