@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import locale
@@ -30,7 +30,6 @@ from .constants import (
     website_url,
 )
 from .fast_data_types import (
-    GLFW_IBEAM_CURSOR,
     GLFW_MOD_ALT,
     GLFW_MOD_SHIFT,
     SingleKey,
@@ -56,8 +55,11 @@ from .utils import (
     cleanup_ssh_control_masters,
     detach,
     expandvars,
+    get_custom_window_icon,
     log_error,
     parse_os_window_state,
+    safe_mtime,
+    shlex_split,
     single_instance,
     startup_notification_handler,
     unix_socket_paths,
@@ -74,7 +76,7 @@ def set_custom_ibeam_cursor() -> None:
     rgba_data2, width2, height2 = load_png_data(data)
     images = (rgba_data, width, height), (rgba_data2, width2, height2)
     try:
-        set_custom_cursor(GLFW_IBEAM_CURSOR, images, 4, 8)
+        set_custom_cursor("beam", images, 4, 8)
     except Exception as e:
         log_error(f'Failed to set custom beam cursor with error: {e}')
 
@@ -170,50 +172,58 @@ def get_macos_shortcut_for(
     return ans
 
 
-def safe_mtime(path: str) -> Optional[float]:
-    with suppress(OSError):
-        return os.path.getmtime(path)
-    return None
-
-
 def set_macos_app_custom_icon() -> None:
-    for name in ('kitty.app.icns', 'kitty.app.png'):
-        icon_path = os.path.join(config_dir, name)
-        custom_icon_mtime = safe_mtime(icon_path)
-        if custom_icon_mtime is not None:
-            from .fast_data_types import cocoa_set_app_icon, cocoa_set_dock_icon
-            krd = getattr(sys, 'kitty_run_data')
-            bundle_path = os.path.dirname(os.path.dirname(krd.get('bundle_exe_dir')))
-            icon_sentinel = os.path.join(bundle_path, 'Icon\r')
-            sentinel_mtime = safe_mtime(icon_sentinel)
-            if sentinel_mtime is None or sentinel_mtime < custom_icon_mtime:
-                try:
-                    cocoa_set_app_icon(icon_path, bundle_path)
-                except (FileNotFoundError, OSError) as e:
-                    log_error(str(e))
-                    log_error('Failed to set custom app icon, ignoring')
-            # macOS Dock does not reload icons until it is restarted, so we set
-            # the application icon here. This will revert when kitty quits, but
-            # can't be helped since there appears to be no way to get the dock
-            # to reload short of killing it.
-            cocoa_set_dock_icon(icon_path)
-            break
+    custom_icon_mtime, custom_icon_path = get_custom_window_icon()
+    if custom_icon_mtime is not None and custom_icon_path is not None:
+        from .fast_data_types import cocoa_set_app_icon, cocoa_set_dock_icon
+        krd = getattr(sys, 'kitty_run_data')
+        bundle_path = os.path.dirname(os.path.dirname(krd.get('bundle_exe_dir')))
+        icon_sentinel = os.path.join(bundle_path, 'Icon\r')
+        sentinel_mtime = safe_mtime(icon_sentinel)
+        if sentinel_mtime is None or sentinel_mtime < custom_icon_mtime:
+            try:
+                cocoa_set_app_icon(custom_icon_path, bundle_path)
+            except (FileNotFoundError, OSError) as e:
+                log_error(str(e))
+                log_error('Failed to set custom app icon, ignoring')
+        # macOS Dock does not reload icons until it is restarted, so we set
+        # the application icon here. This will revert when kitty quits, but
+        # can't be helped since there appears to be no way to get the dock
+        # to reload short of killing it.
+        cocoa_set_dock_icon(custom_icon_path)
+
+
+def get_icon128_path(base_path: str) -> str:
+    # max icon size on X11 64bits is 128x128
+    path, ext = os.path.splitext(base_path)
+    return f'{path}-128{ext}'
 
 
 def set_x11_window_icon() -> None:
-    # max icon size on X11 64bits is 128x128
-    path, ext = os.path.splitext(logo_png_file)
-    set_default_window_icon(f'{path}-128{ext}')
+    custom_icon_path = get_custom_window_icon()[1]
+    try:
+        if custom_icon_path is not None:
+            custom_icon128_path = get_icon128_path(custom_icon_path)
+            if safe_mtime(custom_icon128_path) is None:
+                set_default_window_icon(custom_icon_path)
+            else:
+                set_default_window_icon(custom_icon128_path)
+        else:
+            set_default_window_icon(get_icon128_path(logo_png_file))
+    except ValueError as err:
+        log_error(err)
 
 
-def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+def set_cocoa_global_shortcuts(opts: Options) -> Dict[str, SingleKey]:
     global_shortcuts: Dict[str, SingleKey] = {}
     if is_macos:
         from collections import defaultdict
         func_map = defaultdict(list)
-        for k, v in opts.keymap.items():
-            parts = tuple(v.split())
-            func_map[parts].append(k)
+        for k, v in opts.keyboard_modes[''].keymap.items():
+            for kd in v:
+                if kd.is_suitable_for_global_shortcut:
+                    parts = tuple(kd.definition.split())
+                    func_map[parts].append(k)
 
         for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
                    'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry', 'toggle_fullscreen',
@@ -233,13 +243,20 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ())
         val = get_macos_shortcut_for(func_map, f'open_url {website_url()}', lookup_name='open_kitty_website')
         if val is not None:
             global_shortcuts['open_kitty_website'] = val
+    return global_shortcuts
 
+
+def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+    if is_macos:
+        global_shortcuts = set_cocoa_global_shortcuts(opts)
         if opts.macos_custom_beam_cursor:
             set_custom_ibeam_cursor()
         set_macos_app_custom_icon()
+    else:
+        global_shortcuts = {}
+        if not is_wayland():  # no window icons on wayland
+            set_x11_window_icon()
 
-    if not is_wayland() and not is_macos:  # no window icons on wayland
-        set_x11_window_icon()
     with cached_values_for(run_app.cached_values_name) as cached_values:
         startup_sessions = tuple(create_sessions(opts, args, default_session=opts.startup_session))
         wincls = (startup_sessions[0].os_window_class if startup_sessions else '') or args.cls or appname
@@ -255,8 +272,9 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ())
                     wincls, wstate, load_all_shaders, disallow_override_title=bool(args.title))
         boss = Boss(opts, args, cached_values, global_shortcuts)
         boss.start(window_id, startup_sessions)
-        if bad_lines:
-            boss.show_bad_config_lines(bad_lines)
+        if bad_lines or boss.misc_config_errors:
+            boss.show_bad_config_lines(bad_lines, boss.misc_config_errors)
+            boss.misc_config_errors = []
         try:
             boss.child_monitor.main_loop()
         finally:
@@ -341,17 +359,18 @@ def macos_cmdline(argv_args: List[str]) -> List[str]:
             raw = f.read()
     except FileNotFoundError:
         return argv_args
-    import shlex
     raw = raw.strip()
-    ans = shlex.split(raw)
+    ans = list(shlex_split(raw))
     if ans and ans[0] == 'kitty':
         del ans[0]
     return ans
 
 
 def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
+    if from_config_file and listen_on == 'none':
+        return ''
     listen_on = expandvars(listen_on)
-    if '{kitty_pid}' not in listen_on and from_config_file:
+    if '{kitty_pid}' not in listen_on and from_config_file and listen_on.startswith('unix:'):
         listen_on += '-{kitty_pid}'
     listen_on = listen_on.replace('{kitty_pid}', str(os.getpid()))
     if listen_on.startswith('unix:'):
@@ -362,6 +381,9 @@ def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
             elif not os.path.isabs(path):
                 import tempfile
                 listen_on = f'unix:{os.path.join(tempfile.gettempdir(), path)}'
+    elif listen_on.startswith('tcp:') or listen_on.startswith('tcp6:'):
+        if from_config_file:  # use a random port
+            listen_on = ':'.join(listen_on.split(':', 2)[:2]) + ':0'
     return listen_on
 
 
@@ -424,7 +446,7 @@ def setup_manpath(env: Dict[str, str]) -> None:
 
 def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     from_config_file = False
-    if not cli_opts.listen_on and opts.listen_on.startswith('unix:'):
+    if not cli_opts.listen_on:
         cli_opts.listen_on = opts.listen_on
         from_config_file = True
     if cli_opts.listen_on:
